extends CanvasLayer

@export var list: Control

var config = {}

const configPath = "user://ags.json"
func loadSettings():
	if FileAccess.file_exists(configPath):
		var fa = FileAccess.open(configPath, FileAccess.READ)
		var jstr = fa.get_as_text()
		config = JSON.parse_string(jstr)

	else: config = {}

func saveSettings():
	var fa = FileAccess.open(configPath, FileAccess.WRITE)
	fa.store_string(JSON.stringify(config))
	fa.flush()
	fa.close()

func getEnviroment() -> Environment:
	return get_viewport().world_3d.environment

class SettingSlider extends HBoxContainer:
	var label: Label
	var hslider: HSlider
	var spinbox: SpinBox
	var reset_button: Button

	var key: String
	var default: float
	var _set: Callable
	var minval : float
	var maxval : float
	var step : float
	var on_release : bool
	var _name : String

	func _init(_name: String, _key: String, minval: float, maxval: float, step: float, __set: Callable, _default: float, on_release = false):
		self.key = _key
		self.default = _default
		self._set = __set
		self.minval = minval
		self.maxval = maxval
		self.step = step
		self.on_release = on_release
		self._name = _name
	
	func _ready():
		label = Label.new()
		label.text = _name
		add_child(label)
		label.custom_minimum_size = Vector2i.RIGHT * 256

		hslider = HSlider.new()
		hslider.min_value = minval
		hslider.max_value = maxval
		hslider.step = step
		hslider.size_flags_horizontal = SIZE_EXPAND_FILL
		hslider.size_flags_vertical = SIZE_EXPAND_FILL
		add_child(hslider)

		spinbox = SpinBox.new()
		spinbox.min_value = minval
		spinbox.max_value = maxval
		spinbox.step = step
		add_child(spinbox)

		reset_button = Button.new()
		reset_button.text = "Reset"
		reset_button.pressed.connect(func(): reset())
		add_child(reset_button)
		check_default()

		var val = default
		if key in get_node("/root/AGS").config:
			val = get_node("/root/AGS").config[key]
		else: get_node("/root/AGS").config[key] = val

		hslider.value = val
		spinbox.value = val

		if on_release:
			hslider.drag_ended.connect(func(_changed): var v = hslider.value; spinbox.value = v; get_node("/root/AGS").config[key] = v; get_node("/root/AGS").saveSettings(); _set.call(v); check_default())
			hslider.value_changed.connect(func(v): spinbox.editable = false; spinbox.value = v; spinbox.editable = true)
		else:
			hslider.value_changed.connect(func(v): spinbox.value = v; get_node("/root/AGS").config[key] = v; get_node("/root/AGS").saveSettings(); _set.call(v); check_default())
		spinbox.value_changed.connect(func(v):
			if !spinbox.editable: return ;
			hslider.value = v; get_node("/root/AGS").config[key] = v; get_node("/root/AGS").saveSettings(); _set.call(v); check_default())
		_set.call(val)
	
	func reset():
		hslider.value = default
		spinbox.value = default
		get_node("/root/AGS").config[key] = default
		get_node("/root/AGS").saveSettings()
		_set.call(default)
		check_default()
	
	func check_default():
		var visible = false
		if key in get_node("/root/AGS").config:
			visible = get_node("/root/AGS").config[key] != default
		
		reset_button.modulate = Color.WHITE if visible else Color.TRANSPARENT
		reset_button.disabled = !visible

class SettingCheckbox extends HBoxContainer:
	var box: CheckBox
	var reset_button: Button

	var key: String
	var default: bool
	var _set: Callable
	var _name : String

	func _init(_name: String, _key: String, __set: Callable, _default: bool):
		self.key = _key
		self.default = _default
		self._set = __set
		self._name = _name

	func _ready():
		var val = default
		if key in get_node("/root/AGS").config:
			val = get_node("/root/AGS").config[key]
		else: get_node("/root/AGS").config[key] = val

		var padding = Control.new()
		padding.mouse_filter = MOUSE_FILTER_IGNORE
		padding.custom_minimum_size = Vector2i.RIGHT * 256
		add_child(padding)
		
		box = CheckBox.new()
		box.text = _name
		box.button_pressed = val
		add_child(box)

		reset_button = Button.new()
		reset_button.text = "Reset"
		reset_button.pressed.connect(func(): reset())
		add_child(reset_button)
		check_default()

		box.toggled.connect(func(v): get_node("/root/AGS").config[key] = v; get_node("/root/AGS").saveSettings(); _set.call(v); check_default())
		_set.call(val)
	
	func reset():
		box.button_pressed = default
		get_node("/root/AGS").config[key] = default
		get_node("/root/AGS").saveSettings()
		_set.call(default)
		check_default()
	
	func check_default():
		var visible = false
		if key in get_node("/root/AGS").config:
			visible = get_node("/root/AGS").config[key] != default
		
		reset_button.modulate = Color.WHITE if visible else Color.TRANSPARENT
		reset_button.disabled = !visible

class SettingComboBox extends HBoxContainer:
	var label: Label
	var box: OptionButton
	var reset_button: Button

	var key: String
	var default: int
	var _set: Callable

	func _init(name: String, _key: String, __set: Callable, _default: int):
		self.key = _key
		self.default = _default
		self._set = __set

		label = Label.new()
		label.text = name
		add_child(label)
		label.custom_minimum_size = Vector2i.RIGHT * 256
		
		box = OptionButton.new()
		add_child(box)

	func _ready():
		var val = default
		if key in get_node("/root/AGS").config:
			val = get_node("/root/AGS").config[key]
		else: get_node("/root/AGS").config[key] = val
		box.selected = val

		reset_button = Button.new()
		reset_button.text = "Reset"
		reset_button.pressed.connect(func(): reset())
		add_child(reset_button)
		check_default()

		box.item_selected.connect(func(v): get_node("/root/AGS").config[key] = v; get_node("/root/AGS").saveSettings(); _set.call(v); check_default())
		_set.call(val)
	
	func option(text: String):
		box.add_item(text)
		return self

	func reset():
		box.selected = default
		get_node("/root/AGS").config[key] = default
		get_node("/root/AGS").saveSettings()
		_set.call(default)
		check_default()
	
	func check_default():
		var visible = false
		if key in get_node("/root/AGS").config:
			visible = get_node("/root/AGS").config[key] != default
		
		reset_button.modulate = Color.WHITE if visible else Color.TRANSPARENT
		reset_button.disabled = !visible

func _ready() -> void:
	name = "AGS"
	visible = false

	overrideScript("res://mods/AdvancedGraphicsSettings/World.gd")

	RenderingServer.frame_pre_draw.connect(applySettings)
	get_tree().tree_changed.connect(treeChanged)
	loadSettings()


	var dm = SettingComboBox.new("Debug draw mode", "debug_draw_mode", func(v): get_viewport().debug_draw = v, 0)
	dm.option("Disabled")
	dm.option("Unshaded")
	dm.option("Lighting")
	dm.option("Overdraw")
	dm.option("Wireframe")
	dm.option("Normal buffer")
	dm.option("Voxel GI albedo")
	dm.option("Voxel GI lighting")
	dm.option("Voxel GI emission")
	dm.option("Shadow_atlas")
	dm.option("Directional_shadow_atlas")
	dm.option("Scene_luminance")
	dm.option("Ssao")
	dm.option("Ssil")
	dm.option("Pssm_splits")
	dm.option("Decal_atlas")
	dm.option("Sdfgi")
	dm.option("Sdfgi_probes")
	dm.option("GI buffer")
	dm.option("Disable LOD")
	dm.option("Cluster omni lights")
	dm.option("Cluster spot lights")
	dm.option("Cluster decals")
	dm.option("Cluster reflection probes")
	dm.option("Occluders")
	dm.option("Motion vectors")
	dm.option("Internal buffer")
	list.add_child(dm)

	list.add_child(SettingCheckbox.new("Fullscreen", "fullscreen", func(v): get_viewport().mode = Window.MODE_FULLSCREEN if v else Window.MODE_WINDOWED, true))
	var mons = SettingComboBox.new("Screen", "screen", func(v): get_viewport().current_screen = v; get_viewport().size = DisplayServer.screen_get_size(v), DisplayServer.get_primary_screen())
	for i in range(DisplayServer.get_screen_count()):
		var size = DisplayServer.screen_get_size(i)
		var hz = DisplayServer.screen_get_refresh_rate(i)
		mons.option("Screen " + str(i+1) + " (" + str(size.x) + "x" + str(size.y) + "@" + str(floor(hz*10) / 10) +  "Hz)")
	list.add_child(mons)

	list.add_child(SettingSlider.new("UI scale", "ui_scale", 25, 300, 1, func(v): get_viewport().content_scale_factor = v * 0.01, 100, true))
	list.add_child(SettingSlider.new("Resolution scale", "resolution_scale", 10, 200, 1, func(_v): pass , 100))
	list.add_child(SettingComboBox.new("Upscaling filter", "upscaler", func(_v): pass , 1)
		.option("Bilinear")
		.option("FSR 1.0")
		.option("FSR 2.1"))
	list.add_child(SettingSlider.new("FSR sharpness", "fsr_sharpness", 0, 200, 1, func(v): get_viewport().fsr_sharpness = v * 0.01, 20))
	
	list.add_child(SettingComboBox.new("MSAA", "msaa", func(_v): pass , 1)
		.option("Disabled")
		.option("MSAAx2")
		.option("MSAAx4")
		.option("MSAAx8"))
	list.add_child(SettingCheckbox.new("FXAA", "fxaa", func(v): get_viewport().screen_space_aa = Viewport.SCREEN_SPACE_AA_FXAA if v else Viewport.SCREEN_SPACE_AA_DISABLED, false))
	list.add_child(SettingCheckbox.new("TAA", "taa", func(v): get_viewport().use_taa = v, false))
	list.add_child(SettingCheckbox.new("Debanding", "debanding", func(v): get_viewport().use_debanding = v, true))
	list.add_child(SettingCheckbox.new("Occlusion culling", "occlusion_culling", func(v): get_viewport().use_occlusion_culling = v, true))
	list.add_child(SettingCheckbox.new("Variable Rate Shading", "vrs", func(v): get_viewport().vrs_mode = Viewport.VRS_TEXTURE if v else Viewport.VRS_DISABLED, true))
	list.add_child(SettingSlider.new("Mesh LOD threshold", "mesh_lod_treshold", 0, 10, 0.1, func(v): get_viewport().mesh_lod_threshold = v, 1))

	list.add_child(SettingComboBox.new("Debris", "debris", func(_v): applyEnviromentSettings(), 1)
		.option("Disabled")
		.option("Performance")
		.option("Standard"))
	list.add_child(SettingComboBox.new("Grass", "grass", func(_v): applyEnviromentSettings(), 1)
		.option("Disabled")
		.option("Performance")
		.option("Standard"))

	list.add_child(SettingComboBox.new("Reflections", "ssr", func(_v): pass, 0)
		.option("Disabled")
		.option("Very Low")
		.option("Low")
		.option("Medium")
		.option("High")
		.option("Ultra"))
	list.add_child(SettingComboBox.new("Ambient occlusion", "ssao", func(_v): pass , 2)
		.option("Disabled")
		.option("Very Low")
		.option("Low")
		.option("Medium")
		.option("High")
		.option("Ultra"))
	list.add_child(SettingCheckbox.new("Half-size", "ssao_half_size", func(_v): pass, true))
	list.add_child(SettingComboBox.new("Indirect lighting", "ssil", func(_v): pass , 2)
		.option("Disabled")
		.option("Very Low")
		.option("Low")
		.option("Medium")
		.option("High")
		.option("Ultra"))
	list.add_child(SettingCheckbox.new("Half-size", "ssil_half_size", func(_v): pass, true))
	list.add_child(SettingComboBox.new("SDFGI", "sdfgi", func(_v): pass, 0)
		.option("Disabled")
		.option("Very Low")
		.option("Low")
		.option("Medium")
		.option("High")
		.option("Ultra"))
	list.add_child(SettingComboBox.new("Glow", "glow", func(_v): pass, 1)
		.option("Disabled")
		.option("Low")
		.option("High"))
	list.add_child(SettingComboBox.new("Shadow resolution", "shadow_res", func(_v): pass , 2)
		.option("Very Low")
		.option("Low")
		.option("Medium")
		.option("High")
		.option("Ultra"))
	list.add_child(SettingComboBox.new("Shadow quality", "shadow_quality", func(_v): pass , 3)
		.option("Blocky")
		.option("Very Low")
		.option("Low")
		.option("Medium")
		.option("High")
		.option("Ultra"))
	list.add_child(SettingComboBox.new("Volumetric fog", "volumetric_fog", func(_v): pass , 1)
		.option("Disabled")
		.option("Low")
		.option("High"))

	list.add_child(SettingComboBox.new("Tonemap", "tonemap", func(_v): pass , 3)
		.option("Linear")
		.option("Reinhard")
		.option("Filmic")
		.option("ACES"))
	list.add_child(SettingSlider.new("Tonemap exposure", "tonemap_exposure", 0, 16, 0.1, func(_v): pass, 1))
	list.add_child(SettingSlider.new("Tonemap whitepoint", "tonemap_white", 0, 16, 0.1, func(_v): pass, 0.5))

	list.add_child(SettingSlider.new("Brightness", "brightness", 0.01, 8, 0.1, func(_v): pass, 1))
	list.add_child(SettingSlider.new("Contrast", "contrast", 0.01, 8, 0.1, func(_v): pass, 1))
	list.add_child(SettingSlider.new("Saturation", "stauration", 0.01, 8, 0.1, func(_v): pass, 1))
	
func overrideScript(overrideScriptPath : String):
	var script : Script = load(overrideScriptPath)
	script.reload()
	var parentScript = script.get_base_script();
	script.take_over_path(parentScript.resource_path)

func _unhandled_input(event: InputEvent) -> void:
	if event is InputEventKey && event.is_pressed() && !event.is_echo():
		if event.keycode == KEY_F6:
			visible = !visible
			Input.mouse_mode = Input.MOUSE_MODE_VISIBLE if visible else Input.MOUSE_MODE_CAPTURED

func applySettings():
	if !self.is_node_ready(): return

	get_viewport().scaling_3d_scale = config["resolution_scale"] * 0.01
	get_viewport().scaling_3d_mode = config["upscaler"]
	get_viewport().msaa_3d = config["msaa"]
	
	RenderingServer.directional_shadow_atlas_set_size(int(pow(2, 9 + config["shadow_res"])), true)
	RenderingServer.directional_soft_shadow_filter_set_quality(config["shadow_quality"])

	var env = getEnviroment()
	if !env: return

	var ssao = config["ssao"]
	env.ssao_enabled = ssao > 0
	if env.ssao_enabled:
		RenderingServer.environment_set_ssao_quality(ssao-1, config["ssao_half_size"], 0.5, 4, 0, 64)

	var ssil = config["ssil"]
	env.ssil_enabled = ssil > 0
	if env.ssil_enabled:
		RenderingServer.environment_set_ssil_quality(ssil-1, config["ssil_half_size"], 0.5, 2, 0, 64)

	var ssr = config["ssr"]
	env.ssr_enabled = ssr > 0
	if env.ssr_enabled:
		env.ssr_max_steps = pow(2, ssr)

	var sdfgi = config["sdfgi"]
	env.sdfgi_enabled = sdfgi > 0
	if env.sdfgi_enabled:
		env.sdfgi_cascades = clamp(2 * (sdfgi-1),1,8)

	var glow = config["glow"]
	env.glow_enabled = glow > 0
	RenderingServer.environment_glow_set_use_bicubic_upscale(glow > 1)

	env.volumetric_fog_enabled = config["volumetric_fog"] > 0
	RenderingServer.environment_set_volumetric_fog_filter_active(config["volumetric_fog"] > 1)

	env.tonemap_mode = config["tonemap"]
	env.tonemap_exposure = config["tonemap_exposure"]
	env.tonemap_white = config["tonemap_white"]
	
	env.adjustment_enabled = true
	env.adjustment_brightness = config["brightness"]
	env.adjustment_contrast = config["contrast"]
	env.adjustment_saturation = config["stauration"]

var oldMain : Node = null
func treeChanged():
	if !is_node_ready(): return
	if get_tree() == null: return
	if get_tree().current_scene == oldMain:
		return 
	
	var main = get_tree().current_scene
	oldMain = main
	if !main:
		return
	var mainScript = main.get_script()

	if !mainScript: return
	if mainScript.resource_path == "res://Scripts/Map.gd":
		print("Level changed to ", main.mapName)
		applyEnviromentSettings()

const debrisMaterial = preload("res://Nature/Debris/Files/MT_Debris.tres")
const debrisMaterialAA = preload("res://Nature/Debris/Files/MT_Debris_AA.tres")

const foliageMaterial = preload("res://Nature/Foliage/Files/MT_Foliage.tres")
const foliageMaterialAA = preload("res://Nature/Foliage/Files/MT_Foliage_AA.tres")

const branchesMaterial = preload("res://Nature/Trees/Files/MT_Tree_Branches.tres")
const branchesMaterialAA = preload("res://Nature/Trees/Files/MT_Tree_Branches_AA.tres")

const billboardMaterial = preload("res://Nature/Trees/Files/MT_Tree_Billboard.tres")
const billboardMaterialAA = preload("res://Nature/Trees/Files/MT_Tree_Billboard_AA.tres")

func applyEnviromentSettings():
	if !self.is_node_ready(): return

	var main = get_tree().current_scene
	var mainScript = main.get_script()

	if !mainScript: return
	if mainScript.resource_path != "res://Scripts/Map.gd":
		return
	
	var pebbles = main.get_node_or_null("Content/Spawners/Pebbles")
	var debris = main.get_node_or_null("Content/Spawners/Debris")
	var foliage = main.get_node_or_null("Content/Spawners/Foliage")
	var grass = main.get_node_or_null("Content/Spawners/Grass")
	var trees = main.get_node_or_null("Content/Spawners/Trees")

	var debrisSet = config["debris"]
	var grassSet = config["grass"]

	for p in pebbles.get_children():
		p.visible = debrisSet > 0
	for p in debris.get_children():
		p.visible = debrisSet > 0
		if p.visible:
			p.set_surface_override_material(0, debrisMaterialAA if debrisSet > 1 else debrisMaterial)
	
	for p in foliage.get_children():
		p.visible = grassSet > 0
	for p in grass.get_children():
		p.visible = grassSet > 0
		if p.visible:
			p.set_surface_override_material(0, foliageMaterialAA if grassSet > 1 else foliageMaterial)
	for p in trees.get_children():
		p.set_surface_override_material(0, branchesMaterialAA if grassSet > 1 else branchesMaterial)
		p.set_surface_override_material(1, billboardMaterialAA if grassSet > 1 else billboardMaterial)
